program Exact;

{$APPTYPE CONSOLE}

{$R *.res}

uses
  System.SysUtils, System.Math, Velthuis.BigDecimals, Velthuis.AutoConsole;

procedure Help;
begin
  Writeln;
  Writeln('EXACT');
  Writeln;
  Writeln('Shows the exact values generated for a floating point string, for the');
  Writeln('IEEE-754 binary floating point single,  double and extended precision');
  Writeln('data types, and their hexadecimal representations.');
  Writeln;
  Writeln('Examples:');
  Writeln('  EXACT 1.2345');
  Writeln('  EXACT 17.93e17');
  Writeln('  EXACT 1e-30');
  Writeln;
  Writeln('http://rvelthuis.de/programs/exact.html');
  Writeln;
end;

type
  PUInt16 = ^UInt16;
  PUInt32 = ^UInt32;
  PUInt64 = ^UInt64;

{$POINTERMATH ON}

procedure ExactValue(const S: string);
var
  B: BigDecimal;
  Sgl: Single;
  Dbl: Double;
  Ext: Extended;
  Value: string;
begin
  Writeln('The value:    ', S, ' (', BigDecimal(S).ToString, ')');
  Writeln;

  // Single
  B := S;
  Sgl := Single(B);
  if IsInfinite(Sgl) then
    if Sgl < 0 then
      Value := '-Infinity'
    else
      Value := '+Infinity'
  else
  begin
    B := Sgl;
    Value := B.ToString;
  end;
  Writeln('32 bit float: ', Value);
  Writeln(Format('Hex:          %.8X', [PUInt32(@Sgl)^]));
  Writeln;

  // Double
  B := S;
  Dbl := Double(B);
  if IsInfinite(Dbl) then
    if Dbl < 0 then
      Value := '-Infinity'
    else
      Value := '+Infinity'
  else
  begin
    B := Dbl;
    Value := B.ToString;
  end;
  Writeln('64 bit float: ', Value);
  Writeln(Format('Hex:          %.16X', [PUInt64(@Dbl)^]));
  Writeln;

  {$IF SizeOf(Extended) > SizeOf(Double)}
  // Extended
  B := S;
  Ext := Extended(B);
  if IsInfinite(Ext) then
    if Ext < 0 then
      Value := '-Infinity'
    else
      Value := '+Infinity'
  else
  begin
    B := Ext;
    Value := B.ToString;
  end;
  Writeln('80 bit float: ', Value);
  Writeln(Format('Hex:          %.4X%.16X', [PUInt16(@Ext)[4], PUInt64(@Ext)^]));
  Writeln;
  {$IFEND}
end;

begin
  try
    if ParamCount < 1 then
      Help
    else
      ExactValue(ParamStr(1));
  except
    on E: Exception do
      Writeln(E.ClassName, ': ', E.Message);
  end;
end.
